from time import sleep
from typing import Dict, List

from waldur_client import (
    ComponentUsage,
    SlurmAllocationState,
    WaldurClient,
    WaldurClientException,
)

from waldur_site_agent import common_utils
from waldur_site_agent.slurm_client import logger
from waldur_site_agent.slurm_client import utils as slurm_utils
from waldur_site_agent.slurm_client.exceptions import BackendError
from waldur_site_agent.slurm_client.structures import Allocation

from . import (
    ENABLE_USER_HOMEDIR_ACCOUNT_CREATION,
    USER_AGENT,
    WALDUR_OFFERINGS,
    slurm_backend,
)


def submit_total_usage_for_allocation(
    waldur_rest_client: WaldurClient,
    allocation: Allocation,
    total_usage: Dict[str, float],
    waldur_components,
):
    logger.info("Setting usages: %s", total_usage)
    resource_uuid = allocation.marketplace_uuid
    plan_periods = waldur_rest_client.marketplace_resource_get_plan_periods(
        resource_uuid
    )

    if len(plan_periods) == 0:
        logger.warning(
            "A corresponding ResourcePlanPeriod for allocation %s was not found",
            allocation.name,
        )
        return

    plan_period = plan_periods[0]
    component_types = [component["type"] for component in waldur_components]
    missing_components = set(total_usage) - set(component_types)

    if missing_components:
        logger.warning(
            "The following components are not found in Waldur: %s",
            ", ".join(missing_components),
        )

    usage_objects = [
        ComponentUsage(type=component, amount=amount)
        for component, amount in total_usage.items()
        if component in component_types
    ]
    waldur_rest_client.create_component_usages(plan_period["uuid"], usage_objects)


def sync_allocation_users(
    waldur_rest_client: WaldurClient,
    offering_uuid,
    allocation: Allocation,
    usernames: List[str],
):
    logger.info("Syncing associations")
    # Source of truth - associations in a SLURM cluster
    # The service fetches associations from the cluster and pushes them to Waldur
    associations = waldur_rest_client.list_slurm_associations(
        {"allocation_uuid": allocation.uuid}
    )
    remote_usernames = {association["username"] for association in associations}
    local_usernames = set(usernames)

    stale_usernames = remote_usernames - local_usernames
    common_utils.drop_users_from_allocation(
        waldur_rest_client, allocation, stale_usernames
    )

    new_usernames = local_usernames - remote_usernames
    common_utils.add_users_to_allocation(waldur_rest_client, allocation, new_usernames)

    # Offering users sync
    # The service fetches offering users from Waldur and pushes them to the cluster
    logger.info("Creating associations for offering users")
    offering_users = waldur_rest_client.list_remote_offering_users(
        {
            "offering_uuid": offering_uuid,
        }
    )

    offering_user_usernames = [
        offering_user["username"]
        for offering_user in offering_users
        if offering_user["username"] not in local_usernames
    ]

    common_utils.add_users_to_allocation(
        waldur_rest_client, allocation, offering_user_usernames
    )
    added_users = slurm_backend.add_users_to_account(
        allocation, offering_user_usernames
    )

    if ENABLE_USER_HOMEDIR_ACCOUNT_CREATION:
        slurm_backend.create_user_homedirs(added_users)


def sync_data_from_slurm_to_waldur(
    waldur_rest_client: WaldurClient, offering_uuid, allocation_report
):
    waldur_offering = waldur_rest_client._get_offering(offering_uuid)
    # Push SLURM data to Mastermind using REST client
    for allocation_backend_id, allocation_data in allocation_report.items():
        logger.info("-" * 30)
        try:
            logger.info("Processing %s", allocation_backend_id)
            usernames: List[str] = allocation_data["users"]
            usages: Dict[str, Dict[str, float]] = allocation_data["usage"]
            limits: Dict[str, float] = allocation_data["limits"]

            waldur_resources = waldur_rest_client.filter_marketplace_resources(
                {
                    "backend_id": allocation_backend_id,
                    "offering_uuid": offering_uuid,
                    "state": "OK",
                }
            )
            if len(waldur_resources) == 0:
                logger.warning(
                    "There are no resources in Waldur with backend_id '%s',"
                    "skipping sync",
                    allocation_backend_id,
                )
                continue

            marketplace_resource = waldur_resources[0]
            allocation = Allocation(
                name=marketplace_resource["name"],
                uuid=marketplace_resource["resource_uuid"],
                marketplace_uuid=marketplace_resource["uuid"],
                backend_id=allocation_backend_id,
                project_uuid=marketplace_resource["project_uuid"],
                customer_uuid=marketplace_resource["customer_uuid"],
            )

            # Sync users
            sync_allocation_users(
                waldur_rest_client, offering_uuid, allocation, usernames
            )

            # Submit limits
            if limits is not None:
                limits_str = slurm_utils.prettify_limits(limits)
                logger.info("Setting limits to \n%s", limits_str)

                waldur_rest_client.set_slurm_allocation_limits(
                    allocation.marketplace_uuid, limits
                )

            # Submit usage
            total_usage = usages["TOTAL_ACCOUNT_USAGE"]
            submit_total_usage_for_allocation(
                waldur_rest_client,
                allocation,
                total_usage,
                waldur_offering["components"],
            )
        except WaldurClientException as e:
            logger.exception(
                "Waldur REST client error while processing allocation %s: %s",
                allocation_backend_id,
                e,
            )
        except BackendError as e:
            logger.exception(
                "Waldur SLURM client error while processing allocation %s: %s",
                allocation_backend_id,
                e,
            )


def mark_missing_allocations_as_erred(
    waldur_rest_client: WaldurClient, missing_allocations: List[str]
):
    logger.info("Marking allocations missing in SLURM cluster as ERRED")
    for allocation_info in missing_allocations:
        logger.info("Marking %s allocation as ERRED", allocation_info)
        try:
            waldur_rest_client.set_slurm_allocation_state(
                allocation_info["uuid"], SlurmAllocationState.ERRED
            )
        except WaldurClientException as e:
            logger.exception(
                "Waldur REST client error while marking allocation %s: %s",
                allocation_info["backend_id"],
                e,
            )


def process_offering(offering):
    logger.info("Processing offering %s (%s)", offering["name"], offering["uuid"])
    waldur_rest_client = WaldurClient(
        offering["api_url"], offering["api_token"], USER_AGENT
    )

    waldur_allocations = waldur_rest_client.filter_marketplace_resources(
        {
            "offering_uuid": offering["uuid"],
            "state": "OK",
            "field": ["backend_id", "uuid"],
        }
    )
    waldur_allocation_backend_ids = [
        allocation_data["backend_id"] for allocation_data in waldur_allocations
    ]
    allocation_report = slurm_backend.pull_allocations(waldur_allocation_backend_ids)

    # Allocations existing in Waldur but missing in SLURM cluster
    missing_allocations = [
        account_info
        for account_info in waldur_allocations
        if account_info["backend_id"] not in set(allocation_report.keys())
    ]
    mark_missing_allocations_as_erred(waldur_rest_client, missing_allocations)

    sync_data_from_slurm_to_waldur(
        waldur_rest_client, offering["uuid"], allocation_report
    )


def process_offerings():
    logger.info("Fetching data from SLURM cluster")
    for offering in WALDUR_OFFERINGS:
        try:
            process_offering(offering)
        except Exception as e:
            logger.exception("The application crashed due to the error: %s", e)


def slurm_waldur_sync():
    logger.info("Synching data from SLURM cluster to Waldur")
    while True:
        try:
            process_offerings()
        except Exception as e:
            logger.exception("The application crashed due to the error: %s", e)
        sleep(60 * 60)  # Once per hour



import re
from typing import Dict, List, Set

from . import (
    SLURM_ALLOCATION_NAME_MAX_LEN,
    SLURM_ALLOCATION_PREFIX,
    SLURM_ALLOCATION_REGEX,
    SLURM_CONTAINER_NAME,
    SLURM_CUSTOMER_PREFIX,
    SLURM_DEFAULT_ACCOUNT,
    SLURM_DEPLOYMENT_TYPE,
    SLURM_PROJECT_PREFIX,
    logger,
    utils,
)
from .client import SlurmClient
from .exceptions import BackendError, SlurmError
from .structures import Allocation


class SlurmBackend:
    def __init__(
        self,
    ):
        self.client = SlurmClient(SLURM_DEPLOYMENT_TYPE, SLURM_CONTAINER_NAME)

    def ping(self, raise_exception=False):
        try:
            self.client.list_accounts()
        except SlurmError as err:
            if raise_exception:
                raise err
            logger.info("Error: %s", err)
            return False
        else:
            return True

    def list_tres(self):
        return self.client.list_tres()

    def pull_allocations(self, account_names: List[str]):
        report = {}
        for account_name in account_names:
            try:
                account_report = self.pull_allocation(account_name)
                if account_report is not None:
                    report[account_name] = account_report
            except Exception as e:
                logger.exception(
                    "Error while pulling allocation [%s]: %s", account_name, e
                )
        return report

    def pull_allocation(self, account: str):
        logger.info("Pulling allocation %s", account)
        account_info = self.client.get_account(account)

        if account_info is None:
            logger.warning("There is no %s account in the SLURM cluster", account)
            return None

        users = self.client.list_account_users(account)

        report = self.get_usage_report([account])
        usage = report.get(account)
        if not usage:
            empty_usage = {tres: 0.00 for tres in utils.get_tres_list()}
            usage = {"TOTAL_ACCOUNT_USAGE": empty_usage}

        limits: dict = self.get_allocation_limits(account)
        return {"users": users, "usage": usage, "limits": limits}  # limits can be None

    def get_usage_report(self, accounts):
        """
        Example output:
        {
            "account_name": {
                "TOTAL_ACCOUNT_USAGE": {
                    'cpu': 1,
                    'gres/gpu': 2,
                    'mem': 3,
                },
                "user1": {
                    'cpu': 1,
                    'gres/gpu': 2,
                    'mem': 3,
                },
            }
        }
        """
        report = {}
        lines = self.client.get_usage_report(accounts)

        for line in lines:
            logger.info(
                "Processing usage for %s account and %s user", line.account, line.user
            )
            report.setdefault(line.account, {}).setdefault(line.user, {})
            tres_usage = line.tres_usage
            user_usage_existing = report[line.account][line.user]
            user_usage_new = utils.sum_dicts([user_usage_existing, tres_usage])
            report[line.account][line.user] = user_usage_new

        for account_usage in report.values():
            usages_per_user = account_usage.values()
            total = utils.sum_dicts(usages_per_user)
            account_usage["TOTAL_ACCOUNT_USAGE"] = total

        return report

    def get_allocation_limits(self, account: str):
        lines = self.client.get_resource_limits(account)
        correct_lines = [
            association.tres_limits for association in lines if association.tres_limits
        ]
        if len(correct_lines) > 0:
            limits = {
                tres_name: round(limit) for tres_name, limit in correct_lines[0].items()
            }
            return limits

    def get_allocation_name(self, allocation: Allocation):
        name = allocation.name
        prefix = SLURM_ALLOCATION_PREFIX
        hexpart = allocation.uuid[:5]

        raw_name = "%s%s_%s" % (prefix, hexpart, name)
        incorrect_symbols_regex = r"[^%s]+" % SLURM_ALLOCATION_REGEX
        sanitized_name = re.sub(incorrect_symbols_regex, "", raw_name)
        result_name = sanitized_name[:SLURM_ALLOCATION_NAME_MAX_LEN]
        return result_name.lower()

    def get_project_name(self, name: str):
        return "%s%s" % (SLURM_PROJECT_PREFIX, name)

    def get_customer_name(self, name: str):
        return "%s%s" % (SLURM_CUSTOMER_PREFIX, name)

    def set_allocation_limits(
        self, allocation: Allocation, limits_dict: Dict[str, int]
    ):
        self.client.set_resource_limits(allocation.backend_id, limits_dict)

    def delete_account_safely(self, account: str):
        if self.client.get_account(account):
            logger.info("Deleting account %s from SLURM cluster", account)
            self.client.delete_account(account)
        else:
            logger.warning("There is no account %s in SLURM cluster", account)

    def delete_allocation(
        self,
        allocation: Allocation,
    ):
        account = allocation.backend_id
        project_account = self.get_project_name(allocation.project_uuid)

        if not account.strip():
            raise BackendError("Empty backend_id for allocation: %s" % allocation)

        existing_users = self.client.list_account_users(account)

        self.delete_account_safely(account)

        if (
            len(
                [
                    account
                    for account in self.client.list_accounts()
                    if account.organization == project_account
                    and account.name != project_account
                ]
            )
            == 0
        ):
            self.delete_account_safely(project_account)

        # TODO: delete customer if it doesn't have any associated allocations

        return existing_users

    def create_allocation(
        self,
        allocation: Allocation,
        customer_name: str,
        project_name: str,
        limits: dict,
    ):
        customer_account = self.get_customer_name(allocation.customer_uuid)
        project_account = self.get_project_name(allocation.project_uuid)
        allocation_account = self.get_allocation_name(allocation)

        if not self.client.get_account(customer_account):
            logger.info(
                "Creating SLURM account for customer %s (backend id = %s)",
                customer_name,
                customer_account,
            )
            self.client.create_account(
                name=customer_account,
                description=customer_name,
                organization=customer_account,
            )

        if not self.client.get_account(project_account):
            logger.info(
                "Creating SLURM account for project %s (backend id = %s)",
                project_name,
                project_account,
            )
            self.client.create_account(
                name=project_account,
                description=project_name,
                organization=project_account,
                parent_name=customer_account,
            )

        if self.client.get_account(allocation_account) is not None:
            logger.info(
                "The account %s already exists in the cluster, skipping creation",
                allocation_account,
            )
            return

        logger.info(
            "Creating SLURM account for allocation %s (backend id = %s)",
            allocation.name,
            allocation_account,
        )
        self.client.create_account(
            name=allocation_account,
            description=allocation.name,
            organization=project_account,
        )

        allocation.backend_id = allocation_account

        limits_str = utils.prettify_limits(limits)
        logger.info("Setting limits to: \n%s", limits_str)
        self.set_allocation_limits(allocation, limits)

        return allocation_account

    def sync_users(
        self,
        allocation: Allocation,
        usernames: Set[str],
        all_usernames: Set[str],
    ):
        created_associations = self.add_users_to_account(allocation, usernames)

        all_backend_usernames = self.client.list_account_users(allocation.backend_id)
        backend_usernames = {
            username for username in all_backend_usernames if username in all_usernames
        }
        stale_usernames = backend_usernames - usernames

        removed_associations = self.remove_users_from_account(
            allocation, stale_usernames
        )

        return created_associations, removed_associations

    def add_users_to_account(self, allocation: Allocation, usernames: Set[str]):
        created_associations = []
        for username in usernames:
            try:
                succeeded = self.add_user(allocation, username)
                if succeeded:
                    created_associations.append(username)
            except BackendError as e:
                logger.exception(
                    "Unable to add user %s to account %s, details: %s",
                    username,
                    allocation.backend_id,
                    e,
                )
        return created_associations

    def remove_users_from_account(self, allocation: Allocation, usernames: Set[str]):
        removed_associations = []
        for username in usernames:
            try:
                succeeded = self.remove_user(allocation, username)
                if succeeded:
                    removed_associations.append(username)
            except BackendError as e:
                logger.exception(
                    "Unable to remove user %s from account %s, details: %s",
                    username,
                    allocation.backend_id,
                    e,
                )
        return removed_associations

    def add_user(self, allocation: Allocation, username: str):
        """
        Add association between user and SLURM account if it doesn't exists.
        """
        account = allocation.backend_id

        if not account.strip():
            raise BackendError("Empty backend_id for allocation: %s" % allocation)

        if not self.client.get_association(username, account):
            logger.info("Creating association between %s and %s", username, account)
            try:
                self.client.create_association(username, account, SLURM_DEFAULT_ACCOUNT)
            except SlurmError as err:
                logger.exception("Unable to create association in Slurm: %s", err)
                return False
        return True

    def remove_user(self, allocation, username):
        """
        Delete association between user and SLURM account if it exists.
        """
        account = allocation.backend_id

        if not account.strip():
            raise BackendError("Empty backend_id for allocation: %s" % allocation)

        if self.client.get_association(username, account):
            logger.info("Deleting association between %s and %s", username, account)
            try:
                self.client.delete_association(username, account)
            except SlurmError as err:
                logger.exception("Unable to delete association in Slurm: %s", err)
                return False
        return True

    def create_user_homedirs(self, usernames: Set[str]):
        logger.info("Creating homedirs for users")
        for username in usernames:
            try:
                self.client.create_linux_user_homedir(username)
                logger.info("Homedir for user %s has been created", username)
            except SlurmError as err:
                logger.exception(
                    "Unable to create user homedir for %s, reason: %s",
                    username,
                    err,
                )



import re

from . import base, structures, utils
from .parser import SlurmAssociationLine, SlurmReportLine


class SlurmClient(base.BaseClient):
    """
    This class implements Python client for SLURM.
    See also: https://slurm.schedmd.com/sacctmgr.html
    """

    def list_accounts(self, accounts=None):
        command = ["list", "account"]
        if accounts is not None:
            command.append(",".join(accounts))
        output = self._execute_command(command)
        return [
            self._parse_account(line) for line in output.splitlines() if "|" in line
        ]

    def list_tres(self):
        output = self._execute_command(["list", "tres"])
        return [line.split("|")[0] for line in output.splitlines() if "|" in line]

    def get_account(self, name):
        output = self._execute_command(["show", "account", name])
        lines = [line for line in output.splitlines() if "|" in line]
        if len(lines) == 0:
            return None
        return self._parse_account(lines[0])

    def create_account(self, name, description, organization, parent_name=None):
        parts = [
            "add",
            "account",
            name,
            'description="%s"' % description,
            'organization="%s"' % organization,
        ]
        if parent_name:
            parts.append("parent=%s" % parent_name)
        return self._execute_command(parts)

    def delete_all_users_from_account(self, name):
        return self._execute_command(["remove", "user", "where", "account=%s" % name])

    def account_has_users(self, account):
        output = self._execute_command(
            ["show", "association", "where", "account=%s" % account]
        )
        items = [
            self._parse_association(line) for line in output.splitlines() if "|" in line
        ]
        return any(item.user != "" for item in items)

    def delete_account(self, name):
        if self.account_has_users(name):
            self.delete_all_users_from_account(name)

        return self._execute_command(["remove", "account", "where", "name=%s" % name])

    def set_resource_limits(self, account, limits_dict):
        limits_str = ",".join([f"{key}={value}" for key, value in limits_dict.items()])
        quota = f"GrpTRESMins={limits_str}"
        return self._execute_command(["modify", "account", account, "set", quota])

    def get_association(self, user, account):
        output = self._execute_command(
            ["show", "association", "where", "user=%s" % user, "account=%s" % account]
        )
        lines = [line for line in output.splitlines() if "|" in line]
        if len(lines) == 0:
            return None
        return self._parse_association(lines[0])

    def create_association(self, username, account, default_account=""):
        return self._execute_command(
            [
                "add",
                "user",
                username,
                "account=%s" % account,
                "DefaultAccount=%s" % default_account,
            ]
        )

    def delete_association(self, username, account):
        return self._execute_command(
            [
                "remove",
                "user",
                "where",
                "name=%s" % username,
                "and",
                "account=%s" % account,
            ]
        )

    # TODO: consider 'sshare' or 'sreport' utils
    def get_usage_report(self, accounts):
        month_start, month_end = utils.format_current_month()

        args = [
            "--noconvert",
            "--truncate",
            "--allocations",
            "--allusers",
            "--starttime=%s" % month_start,
            "--endtime=%s" % month_end,
            "--accounts=%s" % ",".join(accounts),
            "--format=Account,ReqTRES,Elapsed,User",
        ]
        output = self._execute_command(args, "sacct", immediate=False)
        return [SlurmReportLine(line) for line in output.splitlines() if "|" in line]

    def get_resource_limits(self, account):
        args = [
            "show",
            "association",
            "format=account,GrpTRESMins",
            "where",
            "accounts=%s" % account,
        ]
        output = self._execute_command(args, immediate=False)
        return [
            SlurmAssociationLine(line) for line in output.splitlines() if "|" in line
        ]

    def list_account_users(self, account):
        args = [
            "list",
            "associations",
            "format=account,user",
            "where",
            "account=%s" % account,
        ]
        output = self._execute_command(args)
        return [
            line.split("|")[1]
            for line in output.splitlines()
            if "|" in line and line[-1] != "|"
        ]

    def create_linux_user_homedir(self, username: str):
        return self._execute_command(
            command_name="/sbin/mkhomedir_helper",
            command=[username],
            immediate=False,
            parsable=False,
        )

    def _parse_account(self, line):
        parts = line.split("|")
        return structures.Account(
            name=parts[0],
            description=parts[1],
            organization=parts[2],
        )

    def _parse_association(self, line):
        parts = line.split("|")
        value = parts[9]
        match = re.match(r"cpu=(\d+)", value)
        if match:
            value = int(match.group(1))
        return structures.Association(
            account=parts[1],
            user=parts[2],
            value=value,
        )

    def _execute_command(
        self, command, command_name="sacctmgr", immediate=True, parsable=True
    ):
        account_command = [command_name]
        if parsable:
            account_command.extend(["--parsable2", "--noheader"])
        if immediate:
            account_command.append("--immediate")
        account_command.extend(command)
        return self.execute_command(account_command)



import abc
import subprocess  # noqa: S404

from . import logger
from .exceptions import ConfigurationError, SlurmError


class BaseClient(metaclass=abc.ABCMeta):
    def __init__(self, slurm_deployment_type, slurm_container_name=None):
        if slurm_deployment_type == "docker":
            if slurm_container_name is None:
                raise ConfigurationError(
                    "Missing name of headnode container for docker-based SLURM"
                )
            self.command_prefix = ["docker", "exec", slurm_container_name]
        else:
            self.command_prefix = []

    def execute_command(self, command):
        final_command = self.command_prefix + command
        try:
            logger.debug("Executing command: %s", " ".join(final_command))
            return subprocess.check_output(  # noqa: S603
                final_command, stderr=subprocess.STDOUT, encoding="utf-8"
            )
        except subprocess.CalledProcessError as e:
            logger.exception('Failed to execute command "%s".', command)
            stdout = e.output or ""
            lines = stdout.splitlines()
            if len(lines) > 0 and lines[0].startswith("Warning: Permanently added"):
                lines = lines[1:]
            stdout = "\n".join(lines)
            raise SlurmError(stdout)



import logging
import os
import sys
import types
from enum import Enum

import yaml


def log_newline(self, how_many_lines=1):
    self.removeHandler(self.console_handler)
    self.addHandler(self.blank_handler)
    for i in range(how_many_lines):
        self.info("")

    self.removeHandler(self.blank_handler)
    self.addHandler(self.console_handler)


console_handler = logging.StreamHandler(sys.stdout)
logger = logging.getLogger(__name__)
formatter = logging.Formatter("[%(levelname)s] [%(asctime)s] %(message)s")
console_handler.setFormatter(formatter)

blank_handler = logging.StreamHandler(sys.stdout)
blank_handler.setLevel(logging.INFO)
blank_handler.setFormatter(logging.Formatter(fmt=""))

logger.addHandler(console_handler)
logger.setLevel(logging.INFO)

logger.console_handler = console_handler
logger.blank_handler = blank_handler
logger.newline = types.MethodType(log_newline, logger)

SLURM_DEPLOYMENT_TYPE = os.environ.get("SLURM_DEPLOYMENT_TYPE", "docker")


class SlurmDeploymentType(Enum):
    NATIVE = "native"
    DOCKER = "docker"


if SLURM_DEPLOYMENT_TYPE not in (
    SlurmDeploymentType.DOCKER.value,
    SlurmDeploymentType.NATIVE.value,
):
    logger.error(
        "SLURM_DEPLOYMENT_TYPE has invalid value: %s. Possible values are %s and %s",
        SLURM_DEPLOYMENT_TYPE,
        SlurmDeploymentType.DOCKER.value,
        SlurmDeploymentType.NATIVE.value,
    )
    exit(1)

SLURM_ALLOCATION_REGEX = "a-zA-Z0-9-_"
SLURM_ALLOCATION_NAME_MAX_LEN = os.environ.get("SLURM_ALLOCATION_NAME_MAX_LEN", 34)

SLURM_CUSTOMER_PREFIX = os.environ.get("SLURM_CUSTOMER_PREFIX", "hpc_")
SLURM_PROJECT_PREFIX = os.environ.get("SLURM_PROJECT_PREFIX", "hpc_")
SLURM_ALLOCATION_PREFIX = os.environ.get("SLURM_ALLOCATION_PREFIX", "hpc_")

SLURM_TRES_CONFIG_PATH = os.environ.get(
    "SLURM_TRES_CONFIG_PATH", "config-components.yaml"
)

with open(SLURM_TRES_CONFIG_PATH, "r") as stream:
    tres_config = yaml.safe_load(stream)
    SLURM_TRES = tres_config


SLURM_DEFAULT_ACCOUNT = os.environ.get("SLURM_DEFAULT_ACCOUNT", "waldur")

SLURM_CONTAINER_NAME = os.environ.get("SLURM_CONTAINER_NAME", "slurmctld")



import pprint

from waldur_client import OfferingComponent, WaldurClient, WaldurClientException

from waldur_site_agent.slurm_client import (
    SLURM_ALLOCATION_NAME_MAX_LEN,
    SLURM_ALLOCATION_PREFIX,
    SLURM_CONTAINER_NAME,
    SLURM_CUSTOMER_PREFIX,
    SLURM_DEFAULT_ACCOUNT,
    SLURM_DEPLOYMENT_TYPE,
    SLURM_PROJECT_PREFIX,
    SLURM_TRES,
)
from waldur_site_agent.slurm_client.exceptions import SlurmError
from waldur_site_agent.slurm_client.structures import Allocation

from . import (
    ENABLE_USER_HOMEDIR_ACCOUNT_CREATION,
    USER_AGENT,
    WALDUR_OFFERINGS,
    WALDUR_SYNC_DIRECTION,
    WaldurSyncDirection,
    logger,
    sentry_dsn,
    slurm_backend,
)


def drop_users_from_allocation(
    waldur_rest_client: WaldurClient, allocation: Allocation, usernames: str
):
    logger.info("Stale usernames: %s", " ,".join(usernames))
    for username in usernames:
        try:
            waldur_rest_client.delete_slurm_association(
                allocation.marketplace_uuid, username
            )
            logger.info(
                "The user %s has been dropped from %s (backend_id: %s)",
                username,
                allocation.name,
                allocation.backend_id,
            )
        except WaldurClientException as e:
            logger.error("User %s can not be dropped due to: %s", username, e)


def add_users_to_allocation(
    waldur_rest_client: WaldurClient, allocation: Allocation, usernames: set
):
    logger.info("New usernames to add to Waldur allocation: %s", " ,".join(usernames))
    for username in usernames:
        try:
            waldur_rest_client.create_slurm_association(
                allocation.marketplace_uuid, username
            )
            logger.info(
                "The user %s has been added to %s (backend_id: %s)",
                username,
                allocation.name,
                allocation.backend_id,
            )
        except WaldurClientException as e:
            logger.error("User %s can not be added due to: %s", username, e)


def create_offering_components():
    for offering in WALDUR_OFFERINGS:
        offering_name = offering["name"]
        offering_uuid = offering["uuid"]
        offering_api_url = offering["api_url"]
        offering_api_token = offering["api_token"]

        logger.info("Processing %s offering", offering_name)
        waldur_rest_client = WaldurClient(
            offering_api_url, offering_api_token, USER_AGENT
        )

        logger.info(
            "Creating offering components data for the following TRES: %s",
            ", ".join(SLURM_TRES.keys()),
        )
        for tres_type, tres_info in SLURM_TRES.items():
            component = OfferingComponent(
                billing_type=tres_info["accounting_type"],
                type=tres_type,
                name=tres_info["label"],
                measured_unit=tres_info["measured_unit"],
                limit_amount=tres_info["limit"],
            )
            try:
                waldur_rest_client.create_offering_component(offering_uuid, component)
            except Exception as e:
                logger.info(
                    "Unable to create a component %s for offering %s (%s):",
                    tres_info["label"],
                    offering_name,
                    offering_uuid,
                )
                logger.exception(e)


def diagnostics():
    logger.info("-" * 10 + "DIAGNOSTICS START" + "-" * 10)
    logger.info("Provided settings:")
    format_string = "{:<30} = {:<10}"
    logger.info(format_string.format("WALDUR_SYNC_DIRECTION", WALDUR_SYNC_DIRECTION))

    for offering in WALDUR_OFFERINGS:
        format_string = "{:<30} = {:<10}"
        offering_uuid = offering["uuid"]
        offering_name = offering["name"]
        offering_api_url = offering["api_url"]
        offering_api_token = offering["api_token"]

        logger.info(format_string.format("Offering name", offering_name))
        logger.info(format_string.format("Offering UUID", offering_uuid))
        logger.info(format_string.format("Waldur API URL", offering_api_url))
        logger.info(format_string.format("SENTRY_DSN", str(sentry_dsn)))

        waldur_rest_client = WaldurClient(
            offering_api_url, offering_api_token, USER_AGENT
        )

        try:
            offering_data = waldur_rest_client.get_marketplace_provider_offering(
                offering_uuid
            )
            logger.info("Offering uuid: %s", offering_data["uuid"])
            logger.info("Offering name: %s", offering_data["name"])
            logger.info("Offering org: %s", offering_data["customer_name"])
            logger.info("Offering state: %s", offering_data["state"])

            logger.info("Offering components:")
            format_string = "{:<10} {:<10} {:<10} {:<10}"
            headers = ["Type", "Name", "Unit", "Limit"]
            logger.info(format_string.format(*headers))
            components = [
                [
                    component["type"],
                    component["name"],
                    component["measured_unit"],
                    component["limit_amount"],
                ]
                for component in offering_data["components"]
            ]
            for component in components:
                logger.info(format_string.format(*component))

            logger.newline()
        except WaldurClientException as err:
            logger.error("Unable to fetch offering data, reason: %s", err)

        logger.newline()
        try:
            orders = waldur_rest_client.list_orders(
                {
                    "offering_uuid": offering_uuid,
                    "state": ["pending-provider", "executing"],
                }
            )
            logger.info("Active orders:")
            format_string = "{:<10} {:<10} {:<10}"
            headers = ["Project", "Type", "State"]
            logger.info(format_string.format(*headers))
            for order in orders:
                logger.info(
                    format_string.format(
                        order["project_name"], order["type"], order["state"]
                    )
                )
        except WaldurClientException as err:
            logger.error("Unable to fetch orders, reason: %s", err)

    format_string = "{:<30} = {:<10}"
    logger.info(format_string.format("SLURM_DEPLOYMENT_TYPE", SLURM_DEPLOYMENT_TYPE))
    logger.info(
        format_string.format(
            "SLURM_ALLOCATION_NAME_MAX_LEN", SLURM_ALLOCATION_NAME_MAX_LEN
        )
    )
    logger.info(format_string.format("SLURM_CUSTOMER_PREFIX", SLURM_CUSTOMER_PREFIX))
    logger.info(format_string.format("SLURM_PROJECT_PREFIX", SLURM_PROJECT_PREFIX))
    logger.info(
        format_string.format("SLURM_ALLOCATION_PREFIX", SLURM_ALLOCATION_PREFIX)
    )
    logger.info(format_string.format("SLURM_DEFAULT_ACCOUNT", SLURM_DEFAULT_ACCOUNT))
    logger.info(format_string.format("SLURM_CONTAINER_NAME", SLURM_CONTAINER_NAME))
    logger.newline()

    logger.info("SLURM tres config file content:\n%s\n", pprint.pformat(SLURM_TRES))

    if WALDUR_SYNC_DIRECTION == WaldurSyncDirection.PULL.value:
        logger.info(
            "Agent is running in pull mode - "
            "pulling orders from Waldur and creating allocations"
        )
    else:
        logger.info("Agent is running in push mode - pushing usage stats to Waldur")

    try:
        slurm_version_info = slurm_backend.client._execute_command(
            ["-V"], "sinfo", immediate=False, parsable=False
        )
        logger.info("Slurm version: %s", slurm_version_info.strip())
    except SlurmError as err:
        logger.error("Unable to fetch SLURM info, reason: %s", err)
        return False

    try:
        slurm_backend.ping(raise_exception=True)
        logger.info("SLURM cluster ping is successful")
    except SlurmError as err:
        logger.error("Unable to ping SLURM cluster, reason: %s", err)

    tres = slurm_backend.list_tres()
    logger.info("Available tres in the cluster: %s", ",".join(tres))

    default_account = slurm_backend.client.get_account(SLURM_DEFAULT_ACCOUNT)
    if default_account is None:
        logger.error("There is no account %s in the cluster", default_account)
        return False
    logger.info('Default parent account "%s" is in place', SLURM_DEFAULT_ACCOUNT)
    logger.newline()

    logger.info("-" * 10 + "DIAGNOSTICS END" + "-" * 10)
    return True


def create_homedirs_for_offering_users():
    if not ENABLE_USER_HOMEDIR_ACCOUNT_CREATION:
        logger.warning(
            "ENABLE_USER_HOMEDIR_ACCOUNT_CREATION disabled, skipping processing"
        )
        return

    for offering in WALDUR_OFFERINGS:
        offering_name = offering["name"]
        offering_uuid = offering["uuid"]
        offering_api_url = offering["api_url"]
        offering_api_token = offering["api_token"]

        logger.info("Creating homedirs for %s offering users", offering_name)

        waldur_rest_client = WaldurClient(
            offering_api_url, offering_api_token, USER_AGENT
        )

        offering_users = waldur_rest_client.list_remote_offering_users(
            {
                "offering_uuid": offering_uuid,
            }
        )

        offering_user_usernames = [
            offering_user["username"] for offering_user in offering_users
        ]

        slurm_backend.create_user_homedirs(offering_user_usernames)



import collections
from dataclasses import dataclass

Account = collections.namedtuple("Account", ["name", "description", "organization"])
Association = collections.namedtuple("Association", ["account", "user", "value"])


@dataclass
class Allocation:
    name: str = ""
    uuid: str = ""
    marketplace_uuid: str = ""
    backend_id: str = ""
    project_uuid: str = ""
    customer_uuid: str = ""



import datetime
from functools import cached_property

from . import utils


def parse_duration(value):
    """
    Returns duration in minutes as a float number.
    For example:
    00:01:00 is equal to 1.0
    00:00:03 is equal to 0.05
    00:01:03 is equal to 1.05
    """
    days_sep = "-"
    us_sep = "."
    simple_fmt = "%H:%M:%S"
    days_fmt = "%d-%H:%M:%S"
    us_fmt = "%H:%M:%S.%f"
    days_us_fmt = "%d-%H:%M:%S.%f"

    if days_sep not in value:
        if us_sep in value:  # Simple time with microseconds
            dt = datetime.datetime.strptime(value, us_fmt)
            delta = datetime.timedelta(
                hours=dt.hour,
                minutes=dt.minute,
                seconds=dt.second,
                microseconds=dt.microsecond,
            )
        else:  # Simple time
            dt = datetime.datetime.strptime(value, simple_fmt)
            delta = datetime.timedelta(
                hours=dt.hour, minutes=dt.minute, seconds=dt.second
            )
    else:
        if us_sep in value:  # Simple time with microseconds and days
            dt = datetime.datetime.strptime(value, days_us_fmt)
            delta = datetime.timedelta(
                days=dt.day,
                hours=dt.hour,
                minutes=dt.minute,
                seconds=dt.second,
                microseconds=dt.microsecond,
            )
        else:  # Simple time with days
            dt = datetime.datetime.strptime(value, days_fmt)
            delta = datetime.timedelta(
                days=dt.day, hours=dt.hour, minutes=dt.minute, seconds=dt.second
            )

    return int(delta.total_seconds()) / 60


class SlurmReportLine:
    def __init__(self, line):
        self._parts = line.split("|")

    @cached_property
    def account(self):
        return self._parts[0].strip()

    @cached_property
    def user(self):
        return self._parts[3]

    @cached_property
    def duration(self):
        return parse_duration(self._parts[2])

    @cached_property
    def _resources(self):
        pairs = self._parts[1].split(",")
        return dict(pair.split("=") for pair in pairs)

    def parse_field(self, field):
        if field not in self._resources:
            return 0
        return utils.parse_int(self._resources[field])

    @cached_property
    def tres_usage(self):
        usage = {}
        tres_list = utils.get_tres_list()
        for resource in self._resources:
            if resource not in tres_list:
                continue
            usage_raw = self.parse_field(resource)
            usage[resource] = usage_raw * self.duration
        if "mem" in usage:
            usage["mem"] = usage["mem"] // 2**20  # Convert from Bytes to MB
        return usage


class SlurmAssociationLine(SlurmReportLine):
    @cached_property
    def user(self):
        return None

    @cached_property
    def duration(self):
        return None

    @cached_property
    def _resources(self):
        if self._parts[1] != "":
            pairs = self._parts[1].split(",")
            return dict(pair.split("=") for pair in pairs)

    @cached_property
    def tres_limits(self):
        resources = self._resources
        if resources:
            return {
                tres: utils.parse_int(limit)
                for tres, limit in resources.items()
                if tres in utils.get_tres_list()
            }



import traceback
from time import sleep

from waldur_client import (
    SlurmAllocationState,
    WaldurClient,
    WaldurClientException,
    is_uuid,
)

from waldur_site_agent.slurm_client import logger
from waldur_site_agent.slurm_client import utils as slurm_utils
from waldur_site_agent.slurm_client.exceptions import BackendError
from waldur_site_agent.slurm_client.structures import Allocation

from . import (
    ENABLE_USER_HOMEDIR_ACCOUNT_CREATION,
    USER_AGENT,
    WALDUR_OFFERINGS,
    common_utils,
    slurm_backend,
)


def create_allocation(waldur_rest_client: WaldurClient, order):
    resource_uuid = order["marketplace_resource_uuid"]
    resource_name = order["resource_name"]
    waldur_allocation_uuid = order["resource_uuid"]
    allocation_limits = slurm_utils.get_tres_limits()

    limit_based_components = set(slurm_utils.get_tres_list()) - set(
        allocation_limits.keys()
    )

    for component_key in limit_based_components:
        allocation_limits[component_key] = order["limits"][component_key]

    logger.info("Creating allocation %s", resource_name)
    resource = waldur_rest_client.get_marketplace_resource(resource_uuid)

    if not is_uuid(resource_uuid):
        logger.error("Unexpected resource UUID format, skipping the order")
        return

    if not is_uuid(waldur_allocation_uuid):
        logger.error("Unexpected allocation UUID format, skipping the order")
        return

    allocation = Allocation(
        name=order["resource_name"],
        uuid=waldur_allocation_uuid,
        marketplace_uuid=resource_uuid,
        project_uuid=order["project_uuid"],
        customer_uuid=order["customer_uuid"],
    )

    if resource["state"] != "Creating":
        logger.info(
            "Setting resource state (%s) to CREATING (current state is %s)",
            resource["uuid"],
            resource["state"],
        )
        waldur_rest_client.set_slurm_allocation_state(
            allocation.marketplace_uuid, SlurmAllocationState.CREATING
        )

    logger.info("Creating account in SLURM cluster")
    slurm_backend.create_allocation(
        allocation,
        project_name=order["project_name"],
        customer_name=order["customer_name"],
        limits=allocation_limits,
    )

    logger.info("Updating allocation metadata in Waldur")
    waldur_rest_client.marketplace_resource_set_backend_id(
        allocation.marketplace_uuid, allocation.backend_id
    )
    waldur_rest_client.set_slurm_allocation_backend_id(
        allocation.marketplace_uuid, allocation.backend_id
    )

    logger.info("Updating allocation limits in Waldur")
    waldur_rest_client.set_slurm_allocation_limits(
        allocation.marketplace_uuid, allocation_limits
    )

    logger.info("Updating order state")
    waldur_rest_client.marketplace_order_set_state_done(order["uuid"])

    logger.info("Updating Waldur allocation state")
    waldur_rest_client.set_slurm_allocation_state(
        allocation.marketplace_uuid, SlurmAllocationState.OK
    )

    return allocation


def add_users_to_allocation(
    waldur_rest_client: WaldurClient,
    resource_uuid,
    allocation: Allocation,
    offering_uuid,
):
    logger.info("Adding users to account in SLURM cluster")

    logger.info("Fetching Waldur resource team")
    team = waldur_rest_client.marketplace_resource_get_team(resource_uuid)
    user_uuids = {user["uuid"] for user in team}

    logger.info("Fetching Waldur offering users")
    offering_users_all = waldur_rest_client.list_remote_offering_users(
        {"offering_uuid": offering_uuid}
    )
    offering_usernames = [
        offering_user["username"]
        for offering_user in offering_users_all
        if offering_user["user_uuid"] in user_uuids
    ]

    logger.info("Adding usernames to account in SLURM cluster")
    added_users = slurm_backend.add_users_to_account(allocation, offering_usernames)
    if ENABLE_USER_HOMEDIR_ACCOUNT_CREATION:
        slurm_backend.create_user_homedirs(added_users)

    common_utils.add_users_to_allocation(waldur_rest_client, allocation, added_users)


def process_order_for_creation(waldur_rest_client: WaldurClient, order: dict):
    # Wait until resource is created
    attempts = 0
    while "marketplace_resource_uuid" not in order:
        if attempts > 4:
            logger.error("Order processing timed out")
            return

        if order["state"] != "executing":
            logger.error("order has unexpected state %s", order["state"])
            return

        logger.info("Waiting for resource creation...")
        sleep(5)

        order = waldur_rest_client.get_order(order["uuid"])
        attempts += 1

    # TODO: drop this cycle...
    # TODO: after removal of waldur_slurm.Allocation model from Mastermind

    while order["resource_uuid"] is None:
        if attempts > 4:
            logger.error("Order processing timed out")
            return

        if order["state"] != "executing":
            logger.error("order has unexpected state %s", order["state"])
            return

        logger.info("Waiting for Waldur allocation creation...")
        sleep(5)

        order = waldur_rest_client.get_order(order["uuid"])
        attempts += 1

    resource_uuid = order["marketplace_resource_uuid"]

    allocation: Allocation = create_allocation(waldur_rest_client, order)

    if allocation is None:
        return

    add_users_to_allocation(
        waldur_rest_client, resource_uuid, allocation, order["offering_uuid"]
    )


def process_order_for_limits_update(waldur_rest_client: WaldurClient, order: dict):
    logger.info("Updating limits for %s", order["resource_name"])
    resource_uuid = order["marketplace_resource_uuid"]
    allocation_uuid = order["resource_uuid"]

    allocation_waldur = waldur_rest_client.get_slurm_allocation(allocation_uuid)
    allocation = Allocation(
        backend_id=allocation_waldur["backend_id"],
        project_uuid=order["project_uuid"],
        customer_uuid=order["customer_uuid"],
    )

    waldur_rest_client.set_slurm_allocation_state(
        resource_uuid, SlurmAllocationState.UPDATING
    )

    limits = order["limits"]
    if not limits:
        logger.error(
            "order %s (allocation %s) with type" + "Update does not include new limits",
            order["uuid"],
            allocation_waldur["name"],
        )

    slurm_backend.set_allocation_limits(allocation, limits)

    logger.info(
        "The limits for %s were updated successfully from %s to %s",
        allocation_waldur["name"],
        order["attributes"]["old_limits"],
        limits,
    )


def process_order_for_termination(waldur_rest_client: WaldurClient, order: dict):
    logger.info("Terminating allocation %s", order["resource_name"])
    allocation_uuid = order["resource_uuid"]

    allocation_waldur = waldur_rest_client.get_slurm_allocation(allocation_uuid)
    allocation = Allocation(
        backend_id=allocation_waldur["backend_id"],
        project_uuid=order["project_uuid"],
        customer_uuid=order["customer_uuid"],
    )
    slurm_backend.delete_allocation(allocation)

    waldur_rest_client.marketplace_order_set_state_done(order["uuid"])
    logger.info("Allocation has been terminated successfully")


def process_offering(offering):
    # Pull data form Mastermind using REST client
    logger.info("Processing offering %s (%s)", offering["name"], offering["uuid"])
    waldur_rest_client = WaldurClient(
        offering["api_url"], offering["api_token"], USER_AGENT
    )

    orders = waldur_rest_client.list_orders(
        {
            "offering_uuid": offering["uuid"],
            "state": ["pending-provider", "executing"],
        }
    )

    if len(orders) == 0:
        logger.info("There are no pending or executing orders")
        return

    for order in orders:
        try:
            logger.info(
                "Processing order %s (%s) with state %s",
                order["attributes"].get("name", "N/A"),
                order["uuid"],
                order["state"],
            )

            if order["state"] == "executing":
                logger.info("Order is executing already, no need for approval")
            else:
                logger.info("Approving the order")
                waldur_rest_client.marketplace_order_approve_by_provider(order["uuid"])
                logger.info("Refreshing the order")
                order = waldur_rest_client.get_order(order["uuid"])

            if order["type"] == "Create":
                process_order_for_creation(waldur_rest_client, order)

            if order["type"] == "Update":
                process_order_for_limits_update(waldur_rest_client, order)

            if order["type"] == "Terminate":
                process_order_for_termination(waldur_rest_client, order)

        except WaldurClientException as e:
            logger.exception(
                "Waldur REST client error while processing order %s: %s",
                order["uuid"],
                e,
            )
        except BackendError as e:
            logger.exception(
                "Waldur SLURM client error while processing order %s: %s",
                order["uuid"],
                e,
            )
            waldur_rest_client.marketplace_order_set_state_erred(
                order["uuid"],
                error_message=str(e),
                error_traceback=traceback.format_exc(),
            )


def process_offerings():
    for offering in WALDUR_OFFERINGS:
        try:
            process_offering(offering)
        except Exception as e:
            logger.exception("The application crashed due to the error: %s", e)


def waldur_slurm_sync():
    logger.info("Synching data from Waldur to SLURM cluster")
    while True:
        logger.info("Pulling data from Waldur to SLURM cluster")
        try:
            process_offerings()
        except Exception as e:
            logger.exception("The application crashed due to the error: %s", e)
        sleep(2 * 60)  # Once per 2 minutes



import os
from enum import Enum
from importlib.metadata import version

import yaml

from waldur_site_agent.slurm_client import logger
from waldur_site_agent.slurm_client.backend import SlurmBackend


# "pull" stands for sync from Waldur to SLURM cluster
# "push" stands for sync from SLURM cluster to Waldur
class WaldurSyncDirection(Enum):
    PULL = "pull"
    PUSH = "push"


waldur_api_url = os.environ.get("WALDUR_API_URL")
waldur_api_token = os.environ.get("WALDUR_API_TOKEN")
waldur_offering_uuid = os.environ.get("WALDUR_OFFERING_UUID")
waldur_offering_name = os.environ.get("WALDUR_OFFERING_NAME", "Single offering")
WALDUR_OFFERINGS = []

if all([waldur_api_url, waldur_api_token, waldur_offering_uuid]):
    logger.info("Using environment variables as a config source")
    WALDUR_OFFERINGS = [
        {
            "name": waldur_offering_name,
            "api_url": waldur_api_url,
            "api_token": waldur_api_token,
            "uuid": waldur_offering_uuid,
        }
    ]
else:
    config_file_path = os.environ.get("WALDUR_CONFIG_FILE_PATH")
    if config_file_path is None:
        logger.error("WALDUR_CONFIG_FILE_PATH variable is missing.")
        exit(1)

    logger.info("Using %s as a config source", config_file_path)

    with open(config_file_path, "r") as stream:
        config = yaml.safe_load(stream)
        offering_list = config["offerings"]
        for offering_info in offering_list:
            WALDUR_OFFERINGS.append(
                {
                    "name": offering_info["name"],
                    "api_url": offering_info["waldur_api_url"],
                    "api_token": offering_info["waldur_api_token"],
                    "uuid": offering_info["waldur_offering_uuid"],
                }
            )


WALDUR_SYNC_DIRECTION = os.environ["WALDUR_SYNC_DIRECTION"]

if WALDUR_SYNC_DIRECTION not in [
    WaldurSyncDirection.PULL.value,
    WaldurSyncDirection.PUSH.value,
]:
    logger.error(
        "SLURM_DEPLOYMENT_TYPE has invalid value: %s. Possible values are %s and %s",
        WALDUR_SYNC_DIRECTION,
        WaldurSyncDirection.PULL.value,
        WaldurSyncDirection.PUSH.value,
    )
    exit(1)


ENABLE_USER_HOMEDIR_ACCOUNT_CREATION = os.environ.get(
    "ENABLE_USER_HOMEDIR_ACCOUNT_CREATION", "false"
)

ENABLE_USER_HOMEDIR_ACCOUNT_CREATION = ENABLE_USER_HOMEDIR_ACCOUNT_CREATION.lower() in [
    "yes",
    "true",
]
waldur_site_agent_version = version("waldur-site-agent")

user_agent_dict = {
    "pull": f"waldur-site-agent-pull/{waldur_site_agent_version}",
    "push": f"waldur-site-agent-push/{waldur_site_agent_version}",
}

USER_AGENT = user_agent_dict.get(WALDUR_SYNC_DIRECTION)


slurm_backend = SlurmBackend()

sentry_dsn = os.environ.get("SENTRY_DSN")

if sentry_dsn:
    import sentry_sdk

    sentry_sdk.init(
        dsn=sentry_dsn,
    )



from waldur_site_agent import logger
from waldur_site_agent.slurm_waldur_utils import slurm_waldur_sync
from waldur_site_agent.waldur_slurm_utils import waldur_slurm_sync

from . import WALDUR_SYNC_DIRECTION, WaldurSyncDirection, waldur_site_agent_version


def main():
    logger.info("Waldur site Agent version: %s, site: SLURM", waldur_site_agent_version)

    logger.info("Running agent in %s mode", WALDUR_SYNC_DIRECTION)
    if WALDUR_SYNC_DIRECTION == WaldurSyncDirection.PULL.value:
        waldur_slurm_sync()
    else:
        slurm_waldur_sync()


if __name__ == "__main__":
    main()



import calendar
import datetime
import re
from typing import Dict

import yaml

from . import SLURM_TRES

UNIT_PATTERN = re.compile(r"(\d+)([KMGTP]?)")

UNITS = {
    "K": 2**10,
    "M": 2**20,
    "G": 2**30,
    "T": 2**40,
}

SLURM_ALLOCATION_REGEX = "a-zA-Z0-9-_"
SLURM_ALLOCATION_NAME_MAX_LEN = 34


def month_start(date):
    return datetime.datetime(day=1, month=date.month, year=date.year)


def month_end(date):
    days_in_month = calendar.monthrange(date.year, date.month)[1]
    return datetime.date(month=date.month, year=date.year, day=days_in_month)


def format_current_month():
    today = datetime.datetime.now()
    start = month_start(today).strftime("%Y-%m-%d")
    end = month_end(today).strftime("%Y-%m-%d")
    return start, end


def sanitize_allocation_name(name):
    incorrect_symbols_regex = r"[^%s]+" % SLURM_ALLOCATION_REGEX
    return re.sub(incorrect_symbols_regex, "", name)


def parse_int(value):
    """
    Convert 5K to 5000.
    """
    match = re.match(UNIT_PATTERN, value)
    if not match:
        return 0
    value = int(match.group(1))
    unit = match.group(2)
    if unit:
        factor = UNITS[unit]
    else:
        factor = 1
    return factor * value


def get_tres_list():
    return SLURM_TRES.keys()


def get_tres_limits():
    return {
        tres: data["limit"]
        for tres, data in SLURM_TRES.items()
        if data["accounting_type"] == "usage"
    }


def sum_dicts(dict_list):
    result_dict = {}

    # Iterate through each dictionary in the list
    for curr_dict in dict_list:
        # Sum values for each key in the current dictionary
        for key, value in curr_dict.items():
            result_dict[key] = result_dict.get(key, 0) + value

    return result_dict


def prettify_limits(limits: Dict[str, int]):
    limits_info = {
        SLURM_TRES[key]["label"]: " ".join(
            [str(value), SLURM_TRES[key]["measured_unit"]]
        )
        for key, value in limits.items()
    }
    return yaml.dump(limits_info)
